/**
 * 
 *   Copyright (c) 2004-2014, Istvan David, Istvan Rath and Daniel Varro
 *   All rights reserved. This program and the accompanying materials
 *   are made available under the terms of the Eclipse Public License v1.0
 *   which accompanies this distribution, and is available at
 *   http://www.eclipse.org/legal/epl-v10.html
 *  
 *   Contributors:
 *   Istvan David - initial API and implementation
 *  
 */
package org.eclipse.viatra.cep.core.engine.runtime;

import java.util.Collection;
import java.util.HashSet;
import java.util.Set;
import org.apache.log4j.Logger;
import org.eclipse.viatra.cep.core.engine.runtime.LatestEventMatch;
import org.eclipse.viatra.cep.core.engine.runtime.util.LatestEventQuerySpecification;
import org.eclipse.viatra.cep.core.metamodels.events.Event;
import org.eclipse.viatra.query.runtime.api.IMatchProcessor;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.ViatraQueryEngine;
import org.eclipse.viatra.query.runtime.api.impl.BaseMatcher;
import org.eclipse.viatra.query.runtime.exception.ViatraQueryException;
import org.eclipse.viatra.query.runtime.matchers.tuple.Tuple;
import org.eclipse.viatra.query.runtime.util.ViatraQueryLoggingUtil;

/**
 * Generated pattern matcher API of the org.eclipse.viatra.cep.core.engine.runtime.latestEvent pattern,
 * providing pattern-specific query methods.
 * 
 * <p>Use the pattern matcher on a given model via {@link #on(ViatraQueryEngine)},
 * e.g. in conjunction with {@link ViatraQueryEngine#on(Notifier)}.
 * 
 * <p>Matches of the pattern will be represented as {@link LatestEventMatch}.
 * 
 * <p>Original source:
 * <code><pre>
 * ////////////EVENT PROCESSING////////////
 * 
 * pattern latestEvent(event : Event) {
 * 	InternalModel.latestEvent(_, event);
 * }
 * </pre></code>
 * 
 * @see LatestEventMatch
 * @see LatestEventProcessor
 * @see LatestEventQuerySpecification
 * 
 */
@SuppressWarnings("all")
public class LatestEventMatcher extends BaseMatcher<LatestEventMatch> {
  /**
   * Initializes the pattern matcher within an existing VIATRA Query engine.
   * If the pattern matcher is already constructed in the engine, only a light-weight reference is returned.
   * The match set will be incrementally refreshed upon updates.
   * @param engine the existing VIATRA Query engine in which this matcher will be created.
   * @throws ViatraQueryException if an error occurs during pattern matcher creation
   * 
   */
  public static LatestEventMatcher on(final ViatraQueryEngine engine) throws ViatraQueryException {
    // check if matcher already exists
    LatestEventMatcher matcher = engine.getExistingMatcher(querySpecification());
    if (matcher == null) {
        matcher = (LatestEventMatcher)engine.getMatcher(querySpecification());
    }
    return matcher;
  }
  
  /**
   * @throws ViatraQueryException if an error occurs during pattern matcher creation
   * @return an initialized matcher
   * @noreference This method is for internal matcher initialization by the framework, do not call it manually.
   * 
   */
  public static LatestEventMatcher create() throws ViatraQueryException {
    return new LatestEventMatcher();
  }
  
  private final static int POSITION_EVENT = 0;
  
  private final static Logger LOGGER = ViatraQueryLoggingUtil.getLogger(LatestEventMatcher.class);
  
  /**
   * Initializes the pattern matcher within an existing VIATRA Query engine.
   * If the pattern matcher is already constructed in the engine, only a light-weight reference is returned.
   * The match set will be incrementally refreshed upon updates.
   * @param engine the existing VIATRA Query engine in which this matcher will be created.
   * @throws ViatraQueryException if an error occurs during pattern matcher creation
   * 
   */
  private LatestEventMatcher() throws ViatraQueryException {
    super(querySpecification());
  }
  
  /**
   * Returns the set of all matches of the pattern that conform to the given fixed values of some parameters.
   * @param pEvent the fixed value of pattern parameter event, or null if not bound.
   * @return matches represented as a LatestEventMatch object.
   * 
   */
  public Collection<LatestEventMatch> getAllMatches(final Event pEvent) {
    return rawGetAllMatches(new Object[]{pEvent});
  }
  
  /**
   * Returns an arbitrarily chosen match of the pattern that conforms to the given fixed values of some parameters.
   * Neither determinism nor randomness of selection is guaranteed.
   * @param pEvent the fixed value of pattern parameter event, or null if not bound.
   * @return a match represented as a LatestEventMatch object, or null if no match is found.
   * 
   */
  public LatestEventMatch getOneArbitraryMatch(final Event pEvent) {
    return rawGetOneArbitraryMatch(new Object[]{pEvent});
  }
  
  /**
   * Indicates whether the given combination of specified pattern parameters constitute a valid pattern match,
   * under any possible substitution of the unspecified parameters (if any).
   * @param pEvent the fixed value of pattern parameter event, or null if not bound.
   * @return true if the input is a valid (partial) match of the pattern.
   * 
   */
  public boolean hasMatch(final Event pEvent) {
    return rawHasMatch(new Object[]{pEvent});
  }
  
  /**
   * Returns the number of all matches of the pattern that conform to the given fixed values of some parameters.
   * @param pEvent the fixed value of pattern parameter event, or null if not bound.
   * @return the number of pattern matches found.
   * 
   */
  public int countMatches(final Event pEvent) {
    return rawCountMatches(new Object[]{pEvent});
  }
  
  /**
   * Executes the given processor on each match of the pattern that conforms to the given fixed values of some parameters.
   * @param pEvent the fixed value of pattern parameter event, or null if not bound.
   * @param processor the action that will process each pattern match.
   * 
   */
  public void forEachMatch(final Event pEvent, final IMatchProcessor<? super LatestEventMatch> processor) {
    rawForEachMatch(new Object[]{pEvent}, processor);
  }
  
  /**
   * Executes the given processor on an arbitrarily chosen match of the pattern that conforms to the given fixed values of some parameters.
   * Neither determinism nor randomness of selection is guaranteed.
   * @param pEvent the fixed value of pattern parameter event, or null if not bound.
   * @param processor the action that will process the selected match.
   * @return true if the pattern has at least one match with the given parameter values, false if the processor was not invoked
   * 
   */
  public boolean forOneArbitraryMatch(final Event pEvent, final IMatchProcessor<? super LatestEventMatch> processor) {
    return rawForOneArbitraryMatch(new Object[]{pEvent}, processor);
  }
  
  /**
   * Returns a new (partial) match.
   * This can be used e.g. to call the matcher with a partial match.
   * <p>The returned match will be immutable. Use {@link #newEmptyMatch()} to obtain a mutable match object.
   * @param pEvent the fixed value of pattern parameter event, or null if not bound.
   * @return the (partial) match object.
   * 
   */
  public LatestEventMatch newMatch(final Event pEvent) {
    return LatestEventMatch.newMatch(pEvent);
  }
  
  /**
   * Retrieve the set of values that occur in matches for event.
   * @return the Set of all values or empty set if there are no matches
   * 
   */
  protected Set<Event> rawAccumulateAllValuesOfevent(final Object[] parameters) {
    Set<Event> results = new HashSet<Event>();
    rawAccumulateAllValues(POSITION_EVENT, parameters, results);
    return results;
  }
  
  /**
   * Retrieve the set of values that occur in matches for event.
   * @return the Set of all values or empty set if there are no matches
   * 
   */
  public Set<Event> getAllValuesOfevent() {
    return rawAccumulateAllValuesOfevent(emptyArray());
  }
  
  @Override
  protected LatestEventMatch tupleToMatch(final Tuple t) {
    try {
        return LatestEventMatch.newMatch((Event) t.get(POSITION_EVENT));
    } catch(ClassCastException e) {
        LOGGER.error("Element(s) in tuple not properly typed!",e);
        return null;
    }
  }
  
  @Override
  protected LatestEventMatch arrayToMatch(final Object[] match) {
    try {
        return LatestEventMatch.newMatch((Event) match[POSITION_EVENT]);
    } catch(ClassCastException e) {
        LOGGER.error("Element(s) in array not properly typed!",e);
        return null;
    }
  }
  
  @Override
  protected LatestEventMatch arrayToMatchMutable(final Object[] match) {
    try {
        return LatestEventMatch.newMutableMatch((Event) match[POSITION_EVENT]);
    } catch(ClassCastException e) {
        LOGGER.error("Element(s) in array not properly typed!",e);
        return null;
    }
  }
  
  /**
   * @return the singleton instance of the query specification of this pattern
   * @throws ViatraQueryException if the pattern definition could not be loaded
   * 
   */
  public static IQuerySpecification<LatestEventMatcher> querySpecification() throws ViatraQueryException {
    return LatestEventQuerySpecification.instance();
  }
}
