/**
 * Copyright (c) 2010-2015, Balazs Grill, Istvan Rath and Daniel Varro
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * Contributors:
 * Balazs Grill - initial API and implementation
 */
package org.eclipse.viatra.query.testing.core;

import com.google.common.base.Joiner;
import com.google.common.base.Objects;
import com.google.common.collect.Maps;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import org.apache.log4j.Appender;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;
import org.apache.log4j.spi.LoggingEvent;
import org.eclipse.emf.common.notify.Notifier;
import org.eclipse.emf.common.util.EList;
import org.eclipse.emf.common.util.TreeIterator;
import org.eclipse.emf.common.util.URI;
import org.eclipse.emf.ecore.EObject;
import org.eclipse.emf.ecore.resource.Resource;
import org.eclipse.emf.ecore.resource.ResourceSet;
import org.eclipse.emf.ecore.resource.impl.ResourceSetImpl;
import org.eclipse.viatra.query.runtime.api.IPatternMatch;
import org.eclipse.viatra.query.runtime.api.IQueryGroup;
import org.eclipse.viatra.query.runtime.api.IQuerySpecification;
import org.eclipse.viatra.query.runtime.api.ViatraQueryEngine;
import org.eclipse.viatra.query.runtime.api.ViatraQueryMatcher;
import org.eclipse.viatra.query.runtime.emf.EMFScope;
import org.eclipse.viatra.query.runtime.util.ViatraQueryLoggingUtil;
import org.eclipse.viatra.query.testing.core.IMatchSetModelProvider;
import org.eclipse.viatra.query.testing.core.InitializedSnapshotMatchSetModelProvider;
import org.eclipse.viatra.query.testing.core.MatchSetRecordDiff;
import org.eclipse.viatra.query.testing.core.SnapshotHelper;
import org.eclipse.viatra.query.testing.core.SnapshotMatchSetModelProvider;
import org.eclipse.viatra.query.testing.core.TestingSeverityAggregatorLogAppender;
import org.eclipse.viatra.query.testing.core.api.JavaObjectAccess;
import org.eclipse.viatra.query.testing.core.api.MatchRecordEquivalence;
import org.eclipse.viatra.query.testing.core.internal.DefaultMatchRecordEquivalence;
import org.eclipse.viatra.query.testing.snapshot.MatchRecord;
import org.eclipse.viatra.query.testing.snapshot.MatchSetRecord;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.CollectionExtensions;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Extension;
import org.eclipse.xtext.xbase.lib.Functions.Function0;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;
import org.eclipse.xtext.xbase.lib.Procedures.Procedure1;
import org.junit.Assert;
import org.junit.Assume;
import org.junit.ComparisonFailure;

/**
 * @author Grill Balazs
 */
@SuppressWarnings("all")
public class ViatraQueryTestCase {
  private final static String SEVERITY_AGGREGATOR_LOGAPPENDER_NAME = (ViatraQueryTestCase.class.getName() + 
    ".severityAggregatorLogAppender");
  
  public final static String UNEXPECTED_MATCH = "Unexpected match";
  
  public final static String EXPECTED_NOT_FOUND = "Expected match not found";
  
  private EMFScope scope = new Function0<EMFScope>() {
    public EMFScope apply() {
      try {
        ResourceSetImpl _resourceSetImpl = new ResourceSetImpl();
        EMFScope _eMFScope = new EMFScope(_resourceSetImpl);
        return _eMFScope;
      } catch (Throwable _e) {
        throw Exceptions.sneakyThrow(_e);
      }
    }
  }.apply();
  
  private boolean isScopeSet = false;
  
  private final List<IMatchSetModelProvider> modelProviders;
  
  @Extension
  private SnapshotHelper snapshotHelper;
  
  private Map<String, JavaObjectAccess> accessMap;
  
  private final TestingSeverityAggregatorLogAppender appender;
  
  public ViatraQueryTestCase() {
    this(Maps.<String, JavaObjectAccess>newHashMap());
  }
  
  public ViatraQueryTestCase(final Map<String, JavaObjectAccess> accessMap) {
    LinkedList<IMatchSetModelProvider> _linkedList = new LinkedList<IMatchSetModelProvider>();
    this.modelProviders = _linkedList;
    this.accessMap = accessMap;
    SnapshotHelper _snapshotHelper = new SnapshotHelper(accessMap);
    this.snapshotHelper = _snapshotHelper;
    Logger _logger = ViatraQueryLoggingUtil.getLogger(ViatraQueryEngine.class);
    final Appender a = _logger.getAppender(
      ViatraQueryTestCase.SEVERITY_AGGREGATOR_LOGAPPENDER_NAME);
    TestingSeverityAggregatorLogAppender _xifexpression = null;
    if ((a instanceof TestingSeverityAggregatorLogAppender)) {
      TestingSeverityAggregatorLogAppender _xblockexpression = null;
      {
        ((TestingSeverityAggregatorLogAppender)a).clear();
        _xblockexpression = ((TestingSeverityAggregatorLogAppender)a);
      }
      _xifexpression = _xblockexpression;
    } else {
      TestingSeverityAggregatorLogAppender _xblockexpression_1 = null;
      {
        final TestingSeverityAggregatorLogAppender na = new TestingSeverityAggregatorLogAppender();
        na.setName(ViatraQueryTestCase.SEVERITY_AGGREGATOR_LOGAPPENDER_NAME);
        Logger _logger_1 = ViatraQueryLoggingUtil.getLogger(ViatraQueryEngine.class);
        _logger_1.addAppender(na);
        _xblockexpression_1 = na;
      }
      _xifexpression = _xblockexpression_1;
    }
    this.appender = _xifexpression;
  }
  
  public void assertLogSeverityThreshold(final Level severity) {
    Level _severity = this.appender.getSeverity();
    int _int = _severity.toInt();
    int _int_1 = severity.toInt();
    boolean _greaterThan = (_int > _int_1);
    if (_greaterThan) {
      Level _severity_1 = this.appender.getSeverity();
      String _string = _severity_1.toString();
      String _plus = (_string + " message on log: ");
      LoggingEvent _event = this.appender.getEvent();
      String _renderedMessage = _event.getRenderedMessage();
      String _plus_1 = (_plus + _renderedMessage);
      Assert.fail(_plus_1);
    }
  }
  
  public void assertLogSeverity(final Level severity) {
    Level _severity = this.appender.getSeverity();
    Assert.assertEquals(severity, _severity);
  }
  
  public boolean loadModel(final URI uri) {
    try {
      boolean _xblockexpression = false;
      {
        final ResourceSetImpl resourceSet = new ResourceSetImpl();
        resourceSet.getResource(uri, true);
        EMFScope _eMFScope = new EMFScope(resourceSet);
        this.scope = _eMFScope;
        _xblockexpression = this.isScopeSet = true;
      }
      return _xblockexpression;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  /**
   * Sets the scope of the VIATRA Query test case
   * 
   * @since 1.5.2
   */
  public boolean setScope(final EMFScope scope) {
    boolean _xblockexpression = false;
    {
      this.scope = scope;
      _xblockexpression = this.isScopeSet = true;
    }
    return _xblockexpression;
  }
  
  public void dispose() {
    final Procedure1<IMatchSetModelProvider> _function = new Procedure1<IMatchSetModelProvider>() {
      @Override
      public void apply(final IMatchSetModelProvider it) {
        it.dispose();
      }
    };
    IterableExtensions.<IMatchSetModelProvider>forEach(this.modelProviders, _function);
    this.modelProviders.clear();
    Set<? extends Notifier> _scopeRoots = this.scope.getScopeRoots();
    for (final Notifier n : _scopeRoots) {
      boolean _matched = false;
      if (!_matched) {
        if (n instanceof ResourceSet) {
          _matched=true;
          EList<Resource> _resources = ((ResourceSet)n).getResources();
          final Procedure1<Resource> _function_1 = new Procedure1<Resource>() {
            @Override
            public void apply(final Resource it) {
              it.unload();
            }
          };
          IterableExtensions.<Resource>forEach(_resources, _function_1);
        }
      }
      if (!_matched) {
        if (n instanceof Resource) {
          _matched=true;
          ((Resource)n).unload();
        }
      }
      if (!_matched) {
        if (n instanceof EObject) {
          _matched=true;
          Resource _eResource = ((EObject)n).eResource();
          _eResource.unload();
        }
      }
    }
  }
  
  public <T extends EObject> void modifyModel(final Class<T> clazz, final Function1<? super T, ? extends Boolean> condition, final Procedure1<? super T> operation) {
    final Function1<IMatchSetModelProvider, Boolean> _function = new Function1<IMatchSetModelProvider, Boolean>() {
      @Override
      public Boolean apply(final IMatchSetModelProvider it) {
        boolean _updatedByModify = it.updatedByModify();
        return Boolean.valueOf((!_updatedByModify));
      }
    };
    final Iterable<IMatchSetModelProvider> nonIncrementals = IterableExtensions.<IMatchSetModelProvider>filter(this.modelProviders, _function);
    CollectionExtensions.<IMatchSetModelProvider>removeAll(this.modelProviders, nonIncrementals);
    final Procedure1<IMatchSetModelProvider> _function_1 = new Procedure1<IMatchSetModelProvider>() {
      @Override
      public void apply(final IMatchSetModelProvider it) {
        it.dispose();
      }
    };
    IterableExtensions.<IMatchSetModelProvider>forEach(nonIncrementals, _function_1);
    final LinkedList<T> elementsToModify = CollectionLiterals.<T>newLinkedList();
    Set<? extends Notifier> _scopeRoots = this.scope.getScopeRoots();
    for (final Notifier n : _scopeRoots) {
      {
        Iterator<? extends Notifier> iterator = null;
        boolean _matched = false;
        if (!_matched) {
          if (n instanceof ResourceSet) {
            _matched=true;
            TreeIterator<Notifier> _allContents = ((ResourceSet)n).getAllContents();
            iterator = _allContents;
          }
        }
        if (!_matched) {
          if (n instanceof Resource) {
            _matched=true;
            TreeIterator<EObject> _allContents = ((Resource)n).getAllContents();
            iterator = _allContents;
          }
        }
        if (!_matched) {
          if (n instanceof EObject) {
            _matched=true;
            TreeIterator<EObject> _eAllContents = ((EObject)n).eAllContents();
            iterator = _eAllContents;
          }
        }
        while (iterator.hasNext()) {
          {
            final Notifier element = iterator.next();
            boolean _isInstance = clazz.isInstance(element);
            if (_isInstance) {
              final T cast = clazz.cast(element);
              Boolean _apply = condition.apply(cast);
              if ((_apply).booleanValue()) {
                T _cast = clazz.cast(element);
                elementsToModify.add(_cast);
              }
            }
          }
        }
      }
    }
    for (final T element : elementsToModify) {
      operation.apply(element);
    }
    this.appender.clear();
  }
  
  public boolean addMatchSetModelProvider(final IMatchSetModelProvider matchSetModelProvider) {
    return this.modelProviders.add(matchSetModelProvider);
  }
  
  public <Match extends IPatternMatch> void assumeMatchSetsAreAvailable(final IQuerySpecification<? extends ViatraQueryMatcher<Match>> querySpecification) {
    final Procedure1<IMatchSetModelProvider> _function = new Procedure1<IMatchSetModelProvider>() {
      @Override
      public void apply(final IMatchSetModelProvider it) {
        try {
          try {
            it.<Match>getMatchSetRecord(ViatraQueryTestCase.this.scope, querySpecification, null);
          } catch (final Throwable _t) {
            if (_t instanceof IllegalArgumentException) {
              final IllegalArgumentException e = (IllegalArgumentException)_t;
              Assume.assumeNoException(e);
            } else {
              throw Exceptions.sneakyThrow(_t);
            }
          }
        } catch (Throwable _e) {
          throw Exceptions.sneakyThrow(_e);
        }
      }
    };
    IterableExtensions.<IMatchSetModelProvider>forEach(this.modelProviders, _function);
  }
  
  public <Match extends IPatternMatch> void assertMatchSetsEqual(final IQuerySpecification<? extends ViatraQueryMatcher<Match>> querySpecification) {
    DefaultMatchRecordEquivalence _defaultMatchRecordEquivalence = new DefaultMatchRecordEquivalence(this.accessMap);
    this.<Match>assertMatchSetsEqual(querySpecification, _defaultMatchRecordEquivalence);
  }
  
  /**
   * Checks if the match sets of the given query specification are equivalent, based on a user specified equivalence logic.
   * 
   * @since 1.6
   */
  public <Match extends IPatternMatch> void assertMatchSetsEqual(final IQuerySpecification<? extends ViatraQueryMatcher<Match>> querySpecification, final MatchRecordEquivalence equivalence) {
    this.validateTestCase();
    final IMatchSetModelProvider reference = IterableExtensions.<IMatchSetModelProvider>head(this.modelProviders);
    Iterable<IMatchSetModelProvider> _tail = IterableExtensions.<IMatchSetModelProvider>tail(this.modelProviders);
    final Procedure1<IMatchSetModelProvider> _function = new Procedure1<IMatchSetModelProvider>() {
      @Override
      public void apply(final IMatchSetModelProvider it) {
        final MatchSetRecordDiff matchDiff = ViatraQueryTestCase.this.<Match>getMatchSetDiff(querySpecification, reference, it, equivalence);
        boolean _isEmpty = matchDiff.isEmpty();
        boolean _not = (!_isEmpty);
        if (_not) {
          final Joiner joiner = Joiner.on("\n");
          StringConcatenation _builder = new StringConcatenation();
          _builder.append("Differences found between reference ");
          Class<? extends IMatchSetModelProvider> _class = reference.getClass();
          String _name = _class.getName();
          _builder.append(_name, "");
          _builder.append(" and match set provider ");
          Class<? extends IMatchSetModelProvider> _class_1 = it.getClass();
          String _name_1 = _class_1.getName();
          _builder.append(_name_1, "");
          Set<MatchRecord> _additions = matchDiff.getAdditions();
          final Function1<MatchRecord, String> _function = new Function1<MatchRecord, String>() {
            @Override
            public String apply(final MatchRecord it) {
              return ViatraQueryTestCase.this.snapshotHelper.prettyPrint(it);
            }
          };
          Iterable<String> _map = IterableExtensions.<MatchRecord, String>map(_additions, _function);
          String _join = joiner.join(_map);
          Set<MatchRecord> _removals = matchDiff.getRemovals();
          final Function1<MatchRecord, String> _function_1 = new Function1<MatchRecord, String>() {
            @Override
            public String apply(final MatchRecord it) {
              return ViatraQueryTestCase.this.snapshotHelper.prettyPrint(it);
            }
          };
          Iterable<String> _map_1 = IterableExtensions.<MatchRecord, String>map(_removals, _function_1);
          String _join_1 = joiner.join(_map_1);
          throw new ComparisonFailure(_builder.toString(), _join, _join_1);
        }
      }
    };
    IterableExtensions.<IMatchSetModelProvider>forEach(_tail, _function);
  }
  
  public void assertMatchSetsEqual(final IQueryGroup queryGroup) {
    DefaultMatchRecordEquivalence _defaultMatchRecordEquivalence = new DefaultMatchRecordEquivalence(this.accessMap);
    this.assertMatchSetsEqual(queryGroup, _defaultMatchRecordEquivalence);
  }
  
  /**
   * Checks if the match sets of the queries contained in the provided query group are equivalent in the scope of added {@link IMatchSetModelProvider} instances.
   * 
   * @since 1.6
   */
  public void assertMatchSetsEqual(final IQueryGroup queryGroup, final MatchRecordEquivalence equivalence) {
    Set<IQuerySpecification<?>> _specifications = queryGroup.getSpecifications();
    final Procedure1<IQuerySpecification<?>> _function = new Procedure1<IQuerySpecification<?>>() {
      @Override
      public void apply(final IQuerySpecification<?> it) {
        ViatraQueryTestCase.this.<IPatternMatch>assertMatchSetsEqual(((IQuerySpecification<? extends ViatraQueryMatcher<IPatternMatch>>) it), equivalence);
      }
    };
    IterableExtensions.<IQuerySpecification<?>>forEach(_specifications, _function);
  }
  
  /**
   * Calculates the differences between the match sets of a given {@link IQuerySpecification} based on the specified {@link IMatchSetModelProvider} instances.
   * 
   * @since 1.6
   */
  public <Match extends IPatternMatch> MatchSetRecordDiff getMatchSetDiff(final IQuerySpecification<? extends ViatraQueryMatcher<Match>> querySpecification, final IMatchSetModelProvider expectedProvider, final IMatchSetModelProvider actualProvider, final MatchRecordEquivalence equivalence) {
    try {
      MatchSetRecordDiff _xblockexpression = null;
      {
        this.validateTestCase();
        Match filter = null;
        MatchSetRecord expected = expectedProvider.<Match>getMatchSetRecord(this.scope, querySpecification, filter);
        MatchRecord _filter = expected.getFilter();
        boolean _notEquals = (!Objects.equal(_filter, null));
        if (_notEquals) {
          MatchRecord _filter_1 = expected.getFilter();
          Match _createMatchForMatchRecord = this.snapshotHelper.<Match>createMatchForMatchRecord(querySpecification, _filter_1);
          filter = _createMatchForMatchRecord;
        }
        final MatchSetRecord actual = actualProvider.<Match>getMatchSetRecord(this.scope, querySpecification, filter);
        MatchRecord _filter_2 = actual.getFilter();
        boolean _notEquals_1 = (!Objects.equal(_filter_2, null));
        if (_notEquals_1) {
          boolean _notEquals_2 = (!Objects.equal(filter, null));
          if (_notEquals_2) {
            throw new IllegalArgumentException(
              ((("Filter is provided by more than one sources: " + expectedProvider) + ", ") + actualProvider));
          } else {
            MatchRecord _filter_3 = actual.getFilter();
            Match _createMatchForMatchRecord_1 = this.snapshotHelper.<Match>createMatchForMatchRecord(querySpecification, _filter_3);
            filter = _createMatchForMatchRecord_1;
            MatchSetRecord _matchSetRecord = expectedProvider.<Match>getMatchSetRecord(this.scope, querySpecification, filter);
            expected = _matchSetRecord;
          }
        }
        _xblockexpression = MatchSetRecordDiff.compute(expected, actual, equivalence);
      }
      return _xblockexpression;
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  public <Match extends IPatternMatch> MatchSetRecordDiff getMatchSetDiff(final IQuerySpecification<? extends ViatraQueryMatcher<Match>> querySpecification, final IMatchSetModelProvider expectedProvider, final IMatchSetModelProvider actualProvider) {
    DefaultMatchRecordEquivalence _defaultMatchRecordEquivalence = new DefaultMatchRecordEquivalence(this.accessMap);
    return this.<Match>getMatchSetDiff(querySpecification, expectedProvider, actualProvider, _defaultMatchRecordEquivalence);
  }
  
  /**
   * Validate the created test configuration before calculating and comparing the query results
   * 
   * @since 1.6
   */
  private void validateTestCase() {
    int _size = this.modelProviders.size();
    boolean _lessThan = (_size < 2);
    if (_lessThan) {
      throw new IllegalArgumentException("At least two model providers shall be set");
    }
    boolean _and = false;
    if (!(!this.isScopeSet)) {
      _and = false;
    } else {
      final Function1<IMatchSetModelProvider, Boolean> _function = new Function1<IMatchSetModelProvider, Boolean>() {
        @Override
        public Boolean apply(final IMatchSetModelProvider it) {
          boolean _or = false;
          if ((it instanceof SnapshotMatchSetModelProvider)) {
            _or = true;
          } else {
            _or = (it instanceof InitializedSnapshotMatchSetModelProvider);
          }
          return Boolean.valueOf(_or);
        }
      };
      boolean _exists = IterableExtensions.<IMatchSetModelProvider>exists(this.modelProviders, _function);
      boolean _not = (!_exists);
      _and = _not;
    }
    if (_and) {
      throw new IllegalArgumentException("Always include a model in the test specification");
    }
  }
}
