/**
 * Copyright (c) 2010-2017, Grill Balázs, IncQueryLabs
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * Contributors:
 * Grill Balázs - initial API and implementation
 */
package org.eclipse.viatra.query.testing.core.coverage;

import com.google.common.base.Objects;
import com.google.common.collect.Iterables;
import com.google.common.collect.Sets;
import com.google.common.math.DoubleMath;
import java.util.Collections;
import java.util.HashMap;
import java.util.Set;
import org.eclipse.viatra.query.runtime.matchers.psystem.PTraceable;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PQueries;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PQuery;
import org.eclipse.viatra.query.testing.core.coverage.CoverageState;
import org.eclipse.xtext.xbase.lib.CollectionLiterals;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.IterableExtensions;

/**
 * Associates {@link CoverageState}s to objects.
 * @param <T> type of objects whose coverages are stored
 * @since 1.6
 */
@SuppressWarnings("all")
public class CoverageInfo<T extends Object> extends HashMap<T, CoverageState> {
  protected final static long serialVersionUID = (-8699692647123679741L);
  
  /**
   * Merge coverage. A key is considered covered, if it's covered in at least one of the infos.
   */
  public CoverageInfo<T> mergeWith(final CoverageInfo<T> other) {
    CoverageInfo<T> result = new CoverageInfo<T>();
    Set<T> _keySet = this.keySet();
    Set<T> _keySet_1 = other.keySet();
    Sets.SetView<T> _union = Sets.<T>union(_keySet, _keySet_1);
    for (final T key : _union) {
      {
        CoverageState state = this.get(key);
        CoverageState otherState = other.get(key);
        if (((state != null) && (otherState != null))) {
          CoverageState _best = state.best(otherState);
          result.put(key, _best);
        } else {
          CoverageState _xifexpression = null;
          boolean _notEquals = (!Objects.equal(state, null));
          if (_notEquals) {
            _xifexpression = state;
          } else {
            _xifexpression = otherState;
          }
          result.put(key, _xifexpression);
        }
      }
    }
    return result;
  }
  
  public double getAggregatedCoveragePercent() {
    double _xblockexpression = (double) 0;
    {
      Set<T> _keySet = this.keySet();
      Iterable<PQuery> _filter = Iterables.<PQuery>filter(_keySet, PQuery.class);
      final Function1<PQuery, Double> _function = new Function1<PQuery, Double>() {
        @Override
        public Double apply(final PQuery it) {
          return Double.valueOf(CoverageInfo.this.getCoveragePercent(it));
        }
      };
      final Iterable<Double> queryCoverages = IterableExtensions.<PQuery, Double>map(_filter, _function);
      double _xifexpression = (double) 0;
      boolean _isEmpty = IterableExtensions.isEmpty(queryCoverages);
      if (_isEmpty) {
        _xifexpression = 0;
      } else {
        _xifexpression = DoubleMath.mean(queryCoverages);
      }
      _xblockexpression = _xifexpression;
    }
    return _xblockexpression;
  }
  
  public double getCoveragePercent(final PQuery query) {
    double _xblockexpression = (double) 0;
    {
      final Set<PTraceable> traceables = PQueries.getTraceables(query);
      final Function1<PTraceable, Boolean> _function = new Function1<PTraceable, Boolean>() {
        @Override
        public Boolean apply(final PTraceable it) {
          CoverageState _get = CoverageInfo.this.get(it);
          return Boolean.valueOf(Objects.equal(_get, CoverageState.COVERED));
        }
      };
      Iterable<PTraceable> _filter = IterableExtensions.<PTraceable>filter(traceables, _function);
      int _size = IterableExtensions.size(_filter);
      final double coveredCount = ((double) _size);
      final Function1<PTraceable, Boolean> _function_1 = new Function1<PTraceable, Boolean>() {
        @Override
        public Boolean apply(final PTraceable it) {
          CoverageState _get = CoverageInfo.this.get(it);
          return Boolean.valueOf(Collections.<CoverageState>unmodifiableSet(CollectionLiterals.<CoverageState>newHashSet(CoverageState.COVERED, CoverageState.NOT_COVERED)).contains(_get));
        }
      };
      Iterable<PTraceable> _filter_1 = IterableExtensions.<PTraceable>filter(traceables, _function_1);
      final int representedCount = IterableExtensions.size(_filter_1);
      double _xifexpression = (double) 0;
      if ((representedCount == 0)) {
        _xifexpression = 0;
      } else {
        _xifexpression = ((coveredCount / representedCount) * 100);
      }
      _xblockexpression = _xifexpression;
    }
    return _xblockexpression;
  }
}
