/**
 * Copyright (c) 2010-2017, Dénes Harmath, IncQueryLabs
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors:
 *   Dénes Harmath - initial API and implementation
 */
package org.eclipse.viatra.query.testing.core.coverage;

import com.google.common.base.Charsets;
import com.google.common.collect.Iterables;
import com.google.common.io.Files;
import java.io.File;
import java.util.List;
import java.util.Set;
import org.eclipse.viatra.query.runtime.api.impl.BaseGeneratedPrivateEMFQuerySpecification;
import org.eclipse.viatra.query.runtime.matchers.psystem.PBody;
import org.eclipse.viatra.query.runtime.matchers.psystem.PConstraint;
import org.eclipse.viatra.query.runtime.matchers.psystem.PTraceable;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PDisjunction;
import org.eclipse.viatra.query.runtime.matchers.psystem.queries.PQuery;
import org.eclipse.viatra.query.testing.core.coverage.CoverageAnalyzer;
import org.eclipse.viatra.query.testing.core.coverage.CoverageInfo;
import org.eclipse.viatra.query.testing.core.coverage.CoverageState;
import org.eclipse.xtend2.lib.StringConcatenation;
import org.eclipse.xtext.xbase.lib.Exceptions;
import org.eclipse.xtext.xbase.lib.Functions.Function1;
import org.eclipse.xtext.xbase.lib.InputOutput;
import org.eclipse.xtext.xbase.lib.IterableExtensions;

/**
 * Utility methods to report the results of a {@link CoverageAnalyzer}.
 * 
 * @since 1.6
 */
@SuppressWarnings("all")
public class CoverageReporter {
  private static CharSequence formatPercent(final double percent) {
    StringConcatenation _builder = new StringConcatenation();
    String _format = String.format("%1$.2f", Double.valueOf(percent));
    _builder.append(_format, "");
    _builder.append("%");
    return _builder;
  }
  
  /**
   * Prints coverage report to the standard output.
   */
  public static void reportConsole(final CoverageAnalyzer it) {
    StringConcatenation _builder = new StringConcatenation();
    {
      CoverageInfo<PTraceable> _coverage = it.getCoverage();
      Set<PTraceable> _keySet = _coverage.keySet();
      Iterable<PQuery> _filter = Iterables.<PQuery>filter(_keySet, PQuery.class);
      final Function1<PQuery, String> _function = new Function1<PQuery, String>() {
        @Override
        public String apply(final PQuery it) {
          return it.getFullyQualifiedName();
        }
      };
      List<PQuery> _sortBy = IterableExtensions.<PQuery, String>sortBy(_filter, _function);
      for(final PQuery query : _sortBy) {
        _builder.append("pattern ");
        String _fullyQualifiedName = query.getFullyQualifiedName();
        _builder.append(_fullyQualifiedName, "");
        _builder.append(" (");
        CoverageInfo<PTraceable> _coverage_1 = it.getCoverage();
        CoverageState _get = _coverage_1.get(query);
        _builder.append(_get, "");
        _builder.append(") ");
        CoverageInfo<PTraceable> _coverage_2 = it.getCoverage();
        double _coveragePercent = _coverage_2.getCoveragePercent(query);
        CharSequence _formatPercent = CoverageReporter.formatPercent(_coveragePercent);
        _builder.append(_formatPercent, "");
        _builder.newLineIfNotEmpty();
        {
          PDisjunction _disjunctBodies = query.getDisjunctBodies();
          Set<PBody> _bodies = _disjunctBodies.getBodies();
          boolean _hasElements = false;
          for(final PBody body : _bodies) {
            if (!_hasElements) {
              _hasElements = true;
            } else {
              _builder.appendImmediate(" or ", "");
            }
            _builder.append("{ (");
            CoverageInfo<PTraceable> _coverage_3 = it.getCoverage();
            CoverageState _get_1 = _coverage_3.get(body);
            _builder.append(_get_1, "");
            _builder.append(")");
            _builder.newLineIfNotEmpty();
            {
              Set<PConstraint> _constraints = body.getConstraints();
              for(final PConstraint constraint : _constraints) {
                _builder.append("\t");
                _builder.append(constraint, "\t");
                _builder.append(" (");
                CoverageInfo<PTraceable> _coverage_4 = it.getCoverage();
                CoverageState _get_2 = _coverage_4.get(constraint);
                _builder.append(_get_2, "\t");
                _builder.append(")");
                _builder.newLineIfNotEmpty();
              }
            }
            _builder.append("}");
            _builder.newLine();
          }
        }
      }
    }
    InputOutput.<String>println(_builder.toString());
  }
  
  /**
   * Saves coverage report in HTML format to the given {@link File}.
   */
  public static void reportHtml(final CoverageAnalyzer it, final File file) {
    try {
      StringConcatenation _builder = new StringConcatenation();
      _builder.append("<html><body>");
      _builder.newLine();
      _builder.append("<h1>Pattern coverage report</h1>");
      _builder.newLine();
      _builder.append("<h2>Overall coverage: ");
      CoverageInfo<PTraceable> _coverage = it.getCoverage();
      double _aggregatedCoveragePercent = _coverage.getAggregatedCoveragePercent();
      CharSequence _formatPercent = CoverageReporter.formatPercent(_aggregatedCoveragePercent);
      _builder.append(_formatPercent, "");
      _builder.append("</h2>");
      _builder.newLineIfNotEmpty();
      _builder.append("<h2>Detailed coverage</h2>");
      _builder.newLine();
      {
        CoverageInfo<PTraceable> _coverage_1 = it.getCoverage();
        Set<PTraceable> _keySet = _coverage_1.keySet();
        Iterable<PQuery> _filter = Iterables.<PQuery>filter(_keySet, PQuery.class);
        final Function1<PQuery, String> _function = new Function1<PQuery, String>() {
          @Override
          public String apply(final PQuery it) {
            return it.getFullyQualifiedName();
          }
        };
        List<PQuery> _sortBy = IterableExtensions.<PQuery, String>sortBy(_filter, _function);
        for(final PQuery query : _sortBy) {
          _builder.append("<details>");
          _builder.newLine();
          _builder.append("\t");
          _builder.append("<summary>");
          _builder.newLine();
          _builder.append("\t\t");
          CoverageState _elvis = null;
          CoverageInfo<PTraceable> _coverage_2 = it.getCoverage();
          CoverageState _get = _coverage_2.get(query);
          if (_get != null) {
            _elvis = _get;
          } else {
            _elvis = CoverageState.NOT_REPRESENTED_UNKNOWN_REASON;
          }
          final CoverageState queryState = _elvis;
          _builder.newLineIfNotEmpty();
          _builder.append("\t\t");
          _builder.append("<span style=\"");
          String _style = CoverageReporter.getStyle(queryState);
          _builder.append(_style, "\t\t");
          _builder.append("\" title=\"");
          String _info = CoverageReporter.getInfo(queryState);
          _builder.append(_info, "\t\t");
          _builder.append("\">");
          _builder.newLineIfNotEmpty();
          _builder.append("\t\t\t");
          {
            List<Object> _publishedAs = query.publishedAs();
            Iterable<BaseGeneratedPrivateEMFQuerySpecification> _filter_1 = Iterables.<BaseGeneratedPrivateEMFQuerySpecification>filter(_publishedAs, BaseGeneratedPrivateEMFQuerySpecification.class);
            boolean _isEmpty = IterableExtensions.isEmpty(_filter_1);
            boolean _not = (!_isEmpty);
            if (_not) {
              _builder.append("private");
            }
          }
          _builder.newLineIfNotEmpty();
          _builder.append("\t\t\t");
          _builder.append("pattern ");
          String _fullyQualifiedName = query.getFullyQualifiedName();
          _builder.append(_fullyQualifiedName, "\t\t\t");
          _builder.append("</span>");
          _builder.newLineIfNotEmpty();
          _builder.append("\t\t");
          CoverageInfo<PTraceable> _coverage_3 = it.getCoverage();
          double _coveragePercent = _coverage_3.getCoveragePercent(query);
          CharSequence _formatPercent_1 = CoverageReporter.formatPercent(_coveragePercent);
          _builder.append(_formatPercent_1, "\t\t");
          _builder.newLineIfNotEmpty();
          _builder.append("\t");
          _builder.append("</summary>");
          _builder.newLine();
          _builder.append("\t");
          int bodyIndex = 0;
          _builder.newLineIfNotEmpty();
          {
            PDisjunction _disjunctBodies = query.getDisjunctBodies();
            Set<PBody> _bodies = _disjunctBodies.getBodies();
            boolean _hasElements = false;
            for(final PBody body : _bodies) {
              if (!_hasElements) {
                _hasElements = true;
              } else {
                _builder.appendImmediate(" or ", "\t");
              }
              _builder.append("\t");
              CoverageState _elvis_1 = null;
              CoverageInfo<PTraceable> _coverage_4 = it.getCoverage();
              CoverageState _get_1 = _coverage_4.get(body);
              if (_get_1 != null) {
                _elvis_1 = _get_1;
              } else {
                _elvis_1 = CoverageState.NOT_REPRESENTED_UNKNOWN_REASON;
              }
              final CoverageState bodyState = _elvis_1;
              _builder.newLineIfNotEmpty();
              _builder.append("\t");
              _builder.append("<div><span style=\"");
              String _style_1 = CoverageReporter.getStyle(bodyState);
              _builder.append(_style_1, "\t");
              _builder.append("\" title=\"");
              String _info_1 = CoverageReporter.getInfo(bodyState);
              _builder.append(_info_1, "\t");
              _builder.append("\">Body #");
              int _plusPlus = bodyIndex++;
              _builder.append(_plusPlus, "\t");
              _builder.append(" {</span>");
              _builder.newLineIfNotEmpty();
              {
                Set<PConstraint> _constraints = body.getConstraints();
                for(final PConstraint constraint : _constraints) {
                  _builder.append("\t");
                  CoverageState _elvis_2 = null;
                  CoverageInfo<PTraceable> _coverage_5 = it.getCoverage();
                  CoverageState _get_2 = _coverage_5.get(constraint);
                  if (_get_2 != null) {
                    _elvis_2 = _get_2;
                  } else {
                    _elvis_2 = CoverageState.NOT_REPRESENTED_UNKNOWN_REASON;
                  }
                  final CoverageState constraintState = _elvis_2;
                  _builder.newLineIfNotEmpty();
                  _builder.append("\t");
                  _builder.append("<div style=\"");
                  String _style_2 = CoverageReporter.getStyle(constraintState);
                  _builder.append(_style_2, "\t");
                  _builder.append("\" title=\"");
                  String _info_2 = CoverageReporter.getInfo(constraintState);
                  _builder.append(_info_2, "\t");
                  _builder.append("\">");
                  _builder.newLineIfNotEmpty();
                  _builder.append("\t");
                  _builder.append(constraint, "\t");
                  _builder.newLineIfNotEmpty();
                  _builder.append("\t");
                  _builder.append("</div>");
                  _builder.newLine();
                }
              }
              _builder.append("\t");
              _builder.append("}</div>");
              _builder.newLine();
            }
          }
          _builder.append("</details>");
          _builder.newLine();
        }
      }
      _builder.append("<h2>Legend</h2>");
      _builder.newLine();
      _builder.append("<p>");
      _builder.newLine();
      _builder.append("\t");
      _builder.append("<span style=\"");
      String _style_3 = CoverageReporter.getStyle(CoverageState.COVERED);
      _builder.append(_style_3, "\t");
      _builder.append("\">Covered</span><br/>");
      _builder.newLineIfNotEmpty();
      _builder.append("\t");
      _builder.append("<span style=\"");
      String _style_4 = CoverageReporter.getStyle(CoverageState.NOT_COVERED);
      _builder.append(_style_4, "\t");
      _builder.append("\">Not covered</span><br/>");
      _builder.newLineIfNotEmpty();
      _builder.append("\t");
      _builder.append("<span style=\"");
      String _style_5 = CoverageReporter.getStyle(CoverageState.NOT_REPRESENTED);
      _builder.append(_style_5, "\t");
      _builder.append("\">Not represented (optimized out)</span><br/>");
      _builder.newLineIfNotEmpty();
      _builder.append("\t");
      _builder.append("<span style=\"");
      String _style_6 = CoverageReporter.getStyle(CoverageState.NOT_REPRESENTED_UNKNOWN_REASON);
      _builder.append(_style_6, "\t");
      _builder.append("\">Not represented by error - please report</span><br/>");
      _builder.newLineIfNotEmpty();
      _builder.append("</p>");
      _builder.newLine();
      _builder.append("</body></html>");
      final String content = _builder.toString();
      Files.write(content, file, Charsets.UTF_8);
    } catch (Throwable _e) {
      throw Exceptions.sneakyThrow(_e);
    }
  }
  
  private static String getStyle(final CoverageState cs) {
    String _switchResult = null;
    if (cs != null) {
      switch (cs) {
        case COVERED:
          _switchResult = "background:springgreen";
          break;
        case NOT_COVERED:
          _switchResult = "background:tomato";
          break;
        case NOT_REPRESENTED:
          _switchResult = "color:lightgray";
          break;
        case NOT_REPRESENTED_UNKNOWN_REASON:
          _switchResult = "color:red";
          break;
        default:
          break;
      }
    }
    return _switchResult;
  }
  
  private static String getInfo(final CoverageState cs) {
    String _switchResult = null;
    if (cs != null) {
      switch (cs) {
        case COVERED:
          _switchResult = "Covered";
          break;
        case NOT_COVERED:
          _switchResult = "Not covered";
          break;
        case NOT_REPRESENTED:
          _switchResult = "Not represented";
          break;
        case NOT_REPRESENTED_UNKNOWN_REASON:
          _switchResult = "Internal error";
          break;
        default:
          break;
      }
    }
    return _switchResult;
  }
}
